@@ -1,28 +1,32 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+
+import logging
+
 import stdnum.exceptions
-from sql import Column, Literal, Null
+from sql import Column, Literal
 from sql.aggregate import Min
 from sql.functions import CharLength
 from stdnum import get_cc_module
+from stdnum.eu.vat import MEMBER_STATES as EU_MEMBER_STATES
 
-from trytond import backend
 from trytond.i18n import gettext
 from trytond.model import (
     DeactivableMixin, Index, ModelSQL, ModelView, MultiValueMixin, Unique,
-    ValueMixin, fields, sequence_ordered)
+    ValueMixin, convert_from, fields, sequence_ordered)
 from trytond.model.exceptions import AccessError
 from trytond.pool import Pool
 from trytond.pyson import Bool, Eval
 from trytond.tools import is_full_text, lstrip_wildcard
-from trytond.tools.multivalue import migrate_property
-from trytond.transaction import Transaction
+from trytond.transaction import Transaction, inactive_records
 from trytond.wizard import Button, StateTransition, StateView, Wizard
 
 from .contact_mechanism import _PHONE_TYPES, _ContactMechanismMixin
 from .exceptions import (
     EraseError, InvalidIdentifierCode, SimilarityWarning, VIESUnavailable)
 
+logger = logging.getLogger(__name__)
+
 
 class Party(
         DeactivableMixin, _ContactMechanismMixin, ModelSQL, ModelView,
@@ -106,15 +110,6 @@ class Party(
                 'readonly': Bool(Eval('replaced_by')),
                 })
 
-    @classmethod
-    def __register__(cls, module_name):
-        super(Party, cls).__register__(module_name)
-
-        table_h = cls.__table_handler__(module_name)
-
-        # Migration from 3.8
-        table_h.not_null_action('name', 'remove')
-
     @staticmethod
     def order_code(tables):
         table, _ = tables[None]
@@ -333,6 +328,86 @@ class Party(
                         return address
         return default_address
 
+    @classmethod
+    def autocomplete(cls, text, domain=None, limit=None, order=None):
+        pool = Pool()
+        Configuration = pool.get('party.configuration')
+        configuration = Configuration(1)
+
+        result = super().autocomplete(
+            text, domain=domain, limit=limit, order=order)
+
+        identifier_types = dict(configuration.get_identifier_types())
+
+        def add(type, code):
+            result.append({
+                    'id': None,
+                    'name': code,
+                    'defaults': {
+                        'identifiers': [{
+                                'type': type,
+                                'code': code,
+                                }],
+                        },
+                    })
+
+        eu_vat = get_cc_module('eu', 'vat')
+        if 'eu_vat' in identifier_types and eu_vat.is_valid(text):
+            code = eu_vat.compact(text)
+            add('eu_vat', code)
+        else:
+            for country in eu_vat.guess_country(text):
+                if 'eu_vat' in identifier_types:
+                    code = eu_vat.compact(country + text)
+                    add('eu_vat', code)
+                elif f'{country}_vat' in identifier_types:
+                    vat = get_cc_module(country, 'vat')
+                    code = vat.compact(text)
+                    add(f'{country}_vat', code)
+        return result
+
+    @classmethod
+    def default_get(cls, fields_names, with_rec_name=True):
+        pool = Pool()
+        Country = pool.get('country.country')
+
+        values = super().default_get(fields_names, with_rec_name=with_rec_name)
+        eu_vat = get_cc_module('eu', 'vat')
+        eu_types = {'eu_vat'} | {f'{m}_vat' for m in EU_MEMBER_STATES}
+        for identifier in values.get('identifiers') or []:
+            if identifier.get('type') in eu_types and identifier.get('code'):
+                code = identifier['code']
+                if identifier['type'] != 'eu_vat':
+                    code = identifier['type'][2:] + code
+                try:
+                    result = eu_vat.check_vies(code)
+                except Exception:
+                    print('exception', code)
+                    logger.debug(
+                        f"Fail to check {identifier['code']}", exc_info=True)
+                    continue
+                if result['valid']:
+                    values['name'] = result['name']
+                    for address in values.get('addresses') or []:
+                        if not address:
+                            break
+                    else:
+                        address = {}
+                        values.setdefault('addresses', []).insert(0, address)
+                    address['street'] = result['address']
+                    country_code = code[:2]
+                    countries = Country.search([
+                            ('code', 'ilike', country_code),
+                            ], limit=1)
+                    if countries:
+                        country, = countries
+                        address['country'] = country.id
+                        if with_rec_name:
+                            address.setdefault('country.', {})['rec_name'] = (
+                                country.rec_name)
+                    break
+        return values
+
 
 class PartyLang(ModelSQL, ValueMixin):
     "Party Lang"
@@ -341,36 +416,6 @@ class PartyLang(ModelSQL, ValueMixin):
         'party.party', "Party", ondelete='CASCADE')
     lang = fields.Many2One('ir.lang', "Language")
 
-    @classmethod
-    def __register__(cls, module_name):
-        pool = Pool()
-        Party = pool.get('party.party')
-        cursor = Transaction().connection.cursor()
-        exist = backend.TableHandler.table_exist(cls._table)
-        table = cls.__table__()
-        party = Party.__table__()
-
-        super(PartyLang, cls).__register__(module_name)
-
-        if not exist:
-            party_h = Party.__table_handler__(module_name)
-            if party_h.column_exist('lang'):
-                query = table.insert(
-                    [table.party, table.lang],
-                    party.select(party.id, party.lang))
-                cursor.execute(*query)
-                party_h.drop_column('lang')
-            else:
-                cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('lang')
-        value_names.append('lang')
-        migrate_property(
-            'party.party', field_names, cls, value_names,
-            parent='party', fields=fields)
-
 
 class PartyCategory(ModelSQL):
     'Party - Category'
@@ -382,6 +427,81 @@ class PartyCategory(ModelSQL):
         'party.category', "Category", ondelete='CASCADE', required=True)
 
 
+IDENTIFIER_VAT = [
+    'ad_nrt',
+    'al_nipt',
+    'ar_cuit',
+    'at_uid',
+    'au_abn',
+    'br_cnpj',
+    'by_unp',
+    'ca_bn',
+    'cl_rut',
+    'cn_uscc',
+    'co_nit',
+    'cr_cpj',
+    'cz_dic',
+    'dk_cvr',
+    'do_rnc',
+    'dz_nif',
+    'ec_ruc',
+    'ee_kmkr',
+    'eg_tn',
+    'es_nif',
+    'fi_alv',
+    'fo_vn',
+    'fr_tva',
+    'gh_tin',
+    'gn_nifp',
+    'gt_nit',
+    'hr_oib',
+    'hu_anum',
+    'id_npwp',
+    'il_hp',
+    'in_gstin',
+    'is_vsk',
+    'it_iva',
+    'jp_cn',
+    'ke_pin',
+    'kr_brn',
+    'lt_pvm',
+    'lu_tva',
+    'lv_pvn',
+    'ma_ice',
+    'mc_tva',
+    'me_pib',
+    'mk_edb',
+    'mx_rfc',
+    'nl_btw',
+    'no_mva',
+    'nz_ird',
+    'pe_ruc',
+    'pl_nip',
+    'pt_nif',
+    'py_ruc',
+    'ro_cf',
+    'rs_pib',
+    'ru_inn',
+    'sg_uen',
+    'si_ddv',
+    'sk_dph',
+    'sm_coe',
+    'sv_nit',
+    'tn_mf',
+    'tr_vkn',
+    'tw_ubn',
+    'uy_rut',
+    've_rif',
+    'vn_mst',
+    ]
+
+
+def replace_vat(code):
+    if code in IDENTIFIER_VAT:
+        code = code.split('_', 1)[0] + '_vat'
+    return code
+
+
 IDENTIFIER_TYPES = [
     ('ad_nrt', "Andorra Tax Number"),
     ('al_nipt', "Albanian VAT Number"),
@@ -394,6 +514,8 @@ IDENTIFIER_TYPES = [
     ('au_abn', "Australian Business Number"),
     ('au_acn', "Australian Company Number"),
     ('au_tfn', "Australian Tax File Number"),
+    ('be_bis', "Belgian BIS Number"),
+    ('be_nn', "Belgian National Number"),
     ('be_vat', "Belgian Enterprise Number"),
     ('bg_egn', "Bulgarian Personal Identity Codes"),
     ('bg_pnf', "Bulgarian Number of a Foreigner"),
@@ -426,22 +548,26 @@ IDENTIFIER_TYPES = [
     ('dk_cvr', "Danish VAT Number"),
     ('do_cedula', "Dominican Republic National Identification Number"),
     ('do_rnc', "Dominican Republic Tax"),
+    ('dz_nif', "Algerian Tax Number"),
     ('ec_ci', "Ecuadorian Personal Identity Code"),
     ('ec_ruc', "Ecuadorian Tax Identification"),
     ('ee_ik', "Estonian Personal ID Number"),
     ('ee_kmkr', "Estonian VAT Number"),
     ('ee_registrikood', "Estonian Organisation Registration Code"),
+    ('eg_tn', "Egyptian Tax Registration Number"),
     ('es_cif', "Spanish Company Tax"),
     ('es_dni', "Spanish Personal Identity Codes"),
     ('es_nie', "Spanish Foreigner Number"),
     ('es_nif', "Spanish VAT Number"),
     ('eu_at_02', "SEPA Identifier of the Creditor (AT-02)"),
+    ('eu_oss', "European VAT on e-Commerce - One Stop Shop"),
     ('eu_vat', "European VAT Number"),
     ('fi_alv', "Finnish VAT Number"),
     ('fi_associationid', "Finnish Association Identifier"),
     ('fi_hetu', "Finnish Personal Identity Code"),
     ('fi_veronumero', "Finnish Individual Tax Number"),
     ('fi_ytunnus', "Finnish Business Identifier"),
+    ('fo_vn', "Faroese Tax Number"),
     ('fr_nif', "French Tax Identification Number"),
     ('fr_nir', "French Personal Identification Number"),
     ('fr_siren', "French Company Identification Number"),
@@ -451,6 +577,8 @@ IDENTIFIER_TYPES = [
         "United Kingdom National Health Service Patient Identifier"),
     ('gb_upn', "English Unique Pupil Number"),
     ('gb_vat', "United Kingdom (and Isle of Man) VAT Number"),
+    ('gh_tin', "Ghanaian Taxpayer Identification Number"),
+    ('gn_nifp', "Guinean Tax Number"),
     ('gr_amka', "Greek Social Security Number"),
     ('gr_vat', "Greek VAT Number"),
     ('gt_nit', "Guatemala Tax Number"),
@@ -462,6 +590,8 @@ IDENTIFIER_TYPES = [
     ('il_hp', "Israeli Company Number"),
     ('il_idnr', "Israeli Identity Number"),
     ('in_aadhaar', "Indian Digital Resident Personal Identity Number"),
+    ('in_epic', "Indian Voter ID"),
+    ('in_gstin', "Indian VAT number"),
     ('in_pan', "Indian Income Tax Identifier"),
     ('is_kennitala',
         "Icelandic Personal and Organisation Identity Code"),
@@ -475,8 +605,10 @@ IDENTIFIER_TYPES = [
     ('lt_pvm', "Lithuanian VAT Number"),
     ('lu_tva', "Luxembourgian VAT Number"),
     ('lv_pvn', "Latvian VAT Number"),
+    ('ma_ice', "Moroccan Tax Number"),
     ('mc_tva', "Monacan VAT Number"),
     ('md_idno', "Moldavian Company Identification Number"),
+    ('mk_edb', "Macedonian Tax Number"),
     ('mt_vat', "Maltese VAT Number"),
     ('mu_nid', "Mauritian National Identifier"),
     ('mx_rfc', "Mexican Tax Number"),
@@ -493,9 +625,11 @@ IDENTIFIER_TYPES = [
     ('nz_ird', "New Zealand Inland Revenue Department Number"),
     ('pe_cui', "Peruvian Identity Number"),
     ('pe_ruc', "Peruvian Company Tax Number"),
+    ('pk_cnic', "Pakistani Computerised National Identity Card Number"),
     ('pl_nip', "Polish VAT Number"),
     ('pl_pesel', "Polish National Identification Number"),
     ('pl_regon', "Polish Register of Economic Units"),
+    ('pt_cc', "Portuguese Identity Number"),
     ('pt_nif', "Portuguese VAT Number"),
     ('py_ruc', "Paraguay Tax Number"),
     ('ro_cf', "Romanian VAT Number"),
@@ -506,10 +640,16 @@ IDENTIFIER_TYPES = [
     ('se_orgnr', "Swedish Company Number"),
     ('se_personnummer', "Swedish Personal Number"),
     ('se_vat', "Swedish VAT Number"),
+    ('si_businessid', "Slovenian Corporate Registration Number"),
     ('si_ddv', "Slovenian VAT Number"),
+    ('si_emso', "Slovenian Unique Master Citizen Number"),
     ('sk_dph', "Slovak VAT Number"),
     ('sk_rc', "Slovak Birth Number"),
     ('sm_coe', "San Marino National Tax Number"),
+    ('th_moa', "Thai Memorandum of Association Number"),
+    ('th_pin', "Thai Personal Identification Number"),
+    ('th_tin', "Thai Taxpayer Identification Number"),
+    ('tn_mf', "Tunisian Tax Number"),
     ('tr_tckimlik', "Turkish Personal Identification Number"),
     ('ua_edrpou', "Ukrainian Identifier for Enterprises and Organizations"),
     ('ua_rntrc', "Ukrainian Individual Taxpayer Registration Number"),
@@ -526,6 +666,10 @@ IDENTIFIER_TYPES = [
     ('za_tin', "South African Tax Identification Number"),
     ]
 
+IDENTIFIER_TYPES = list(map(
+        lambda x: (replace_vat(x[0]), x[1]),
+        IDENTIFIER_TYPES))
+
 TAX_IDENTIFIER_TYPES = [
     'ad_nrt',
     'al_nipt',
@@ -545,19 +689,25 @@ TAX_IDENTIFIER_TYPES = [
     'de_vat',
     'dk_cvr',
     'do_rnc',
+    'dz_nif',
     'ec_ruc',
     'ee_kmkr',
+    'eg_tn',
     'es_nif',
     'eu_vat',
     'fi_alv',
+    'fo_vn',
     'fr_tva',
     'gb_vat',
+    'gh_tin',
+    'gn_nifp',
     'gr_vat',
     'gt_nit',
     'hu_anum',
     'id_npwp',
     'ie_vat',
     'il_hp',
+    'in_gstin',
     'is_vsk',
     'it_iva',
     'jp_cn',
@@ -565,8 +715,10 @@ TAX_IDENTIFIER_TYPES = [
     'lt_pvm',
     'lu_tva',
     'lv_pvn',
+    'ma_ice',
     'mc_tva',
     'md_idno',
+    'mk_edb',
     'mt_vat',
     'mx_rfc',
     'nl_btw',
@@ -583,6 +735,8 @@ TAX_IDENTIFIER_TYPES = [
     'si_ddv',
     'sk_dph',
     'sm_coe',
+    'th_tin',
+    'tn_mf',
     'ua_edrpou',
     'ua_rntrc',
     'us_atin',
@@ -597,6 +751,8 @@ TAX_IDENTIFIER_TYPES = [
     'za_tin',
     ]
 
+TAX_IDENTIFIER_TYPES = list(map(replace_vat, TAX_IDENTIFIER_TYPES))
+
 
 class Identifier(sequence_ordered(), DeactivableMixin, ModelSQL, ModelView):
     'Party Identifier'
@@ -623,42 +779,22 @@ class Identifier(sequence_ordered(), DeactivableMixin, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        pool = Pool()
-        Party = pool.get('party.party')
         cursor = Transaction().connection.cursor()
-        party = Party.__table__()
         table = cls.__table__()
 
         super().__register__(module_name)
 
-        party_h = Party.__table_handler__(module_name)
-        if (party_h.column_exist('vat_number')
-                and party_h.column_exist('vat_country')):
-            identifiers = []
-            cursor.execute(*party.select(
-                    party.id, party.vat_number, party.vat_country,
-                    where=(party.vat_number != Null)
-                    | (party.vat_country != Null)))
-            for party_id, number, country in cursor:
-                code = (country or '') + (number or '')
-                if not code:
-                    continue
-                for type in Party.tax_identifier_types():
-                    module = get_cc_module(*type.split('_', 1))
-                    if module.is_valid(code):
-                        break
-                else:
-                    type = None
-                identifiers.append(
-                    cls(party=party_id, code=code, type=type))
-            cls.save(identifiers)
-            party_h.drop_column('vat_number')
-            party_h.drop_column('vat_country')
-
         # Migration from 5.8: Rename cn_rit into cn_ric
         cursor.execute(*table.update([table.type], ['cn_ric'],
                 where=(table.type == 'cn_rit')))
 
+        # Migration from 6.8: Use vat alias
+        for old in IDENTIFIER_VAT:
+            new = replace_vat(old)
+            cursor.execute(*table.update(
+                    [table.type], [new],
+                    where=table.type == old))
+
     @classmethod
     def get_types(cls):
         pool = Pool()
@@ -710,6 +846,30 @@ class Identifier(sequence_ordered(), DeactivableMixin, ModelSQL, ModelView):
                             code=self.code,
                             party=party))
 
+    @fields.depends(methods=['_notify_duplicate'])
+    def on_change_notify(self):
+        notifications = super().on_change_notify()
+        notifications.extend(self._notify_duplicate())
+        return notifications
+
+    @fields.depends('code', methods=['on_change_with_code'])
+    def _notify_duplicate(self):
+        cls = self.__class__
+        if self.code:
+            code = self.on_change_with_code()
+            others = cls.search([
+                    ('id', '!=', self.id),
+                    ('type', '!=', None),
+                    ('code', '=', code),
+                    ], limit=1)
+            if others:
+                other, = others
+                yield ('warning', gettext(
+                        'party.msg_party_identifier_duplicate',
+                        party=other.party.rec_name,
+                        type=other.type_string,
+                        code=other.code))
+
 
 class CheckVIESResult(ModelView):
     'Check VIES'
@@ -913,21 +1073,9 @@ class Erase(Wizard):
         Party = pool.get('party.party')
         cursor = Transaction().connection.cursor()
 
-        def convert_from(table, tables):
-            right, condition = tables[None]
-            if table:
-                table = table.join(right, condition=condition)
-            else:
-                table = right
-            for k, sub_tables in tables.items():
-                if k is None:
-                    continue
-                table = convert_from(table, sub_tables)
-            return table
-
         resources = self.get_resources()
         parties = replacing = [self.ask.party]
-        with Transaction().set_context(active_test=False):
+        with inactive_records():
             while replacing:
                 replacing = Party.search([
                         ('replaced_by', 'in', list(map(int, replacing))),
@@ -944,7 +1092,7 @@ class Erase(Wizard):
                     Model.write(records, {'active': False})
 
                 tables, where = Model.search_domain(domain, active_test=False)
-                from_ = convert_from(None, tables)
+                from_ = convert_from(None, tables, type_='INNER')
                 table, _ = tables[None]
                 query = from_.select(table.id, where=where)
 
